// Learn TypeScript:
//  - https://docs.cocos.com/creator/manual/en/scripting/typescript.html
// Learn Attribute:
//  - https://docs.cocos.com/creator/manual/en/scripting/reference/attributes.html
// Learn life-cycle callbacks:
//  - https://docs.cocos.com/creator/manual/en/scripting/life-cycle-callbacks.html

const {ccclass, property} = cc._decorator;

@ccclass
export default class HotUpdate extends cc.Component {

    @property(cc.Asset)
    manifestUrl:cc.Asset = null;             //更新文件

    private checkBtn:cc.Node = null;         //检查更新
    private updateBtn:cc.Node = null;        //立即更新
    private retryBtn:cc.Node = null;         //重试更新
    private msgLab:cc.Label = null;          //信息显示
    private fileLabel:cc.Label = null;       //文件进度
    private byteLabel:cc.Label = null;       //字节进度
    private fileProgress:cc.ProgressBar = null;   //文件进度
    private byteProgress:cc.ProgressBar = null;   //字节进度
    private versionLab:cc.Label;                  //版本文本

    private versionCompareHandle = null;
    private _storagePath:string;     
    private _am = null;
    private _checkListener = null;
    private _updating = false;
    private _canRetry = false;
    private _updateListener = null;
    private customManifestStr:string ;
    private _failCount = 0;


    onLoad(){
        this.checkBtn = cc.find("update/updatePanel/checkBtn", this.node);
        this.updateBtn = cc.find("update/updatePanel/updateBtn", this.node);
        this.retryBtn = cc.find("update/updatePanel/retryBtn", this.node);
        this.msgLab = cc.find("update/updatePanel/msgLab", this.node).getComponent(cc.Label);
        this.fileLabel = cc.find("update/updatePanel/fileLabel", this.node).getComponent(cc.Label);
        this.byteLabel = cc.find("update/updatePanel/byteLabel", this.node).getComponent(cc.Label);
        this.fileProgress = cc.find("update/updatePanel/fileProgress", this.node).getComponent(cc.ProgressBar);
        this.byteProgress = cc.find("update/updatePanel/byteProgress", this.node).getComponent(cc.ProgressBar);
        this.versionLab = cc.find("update/updatePanel/versionLab", this.node).getComponent(cc.Label);

        // Hot update is only available in Native build
        console.log("isNative:", cc.sys.isNative);
        if (!cc.sys.isNative) {
            return;
        }

        this.checkBtn.on(cc.Node.EventType.TOUCH_END, this.checkUpdate, this);
        this.updateBtn.on(cc.Node.EventType.TOUCH_END, this.hotUpdate, this);
        this.retryBtn.on(cc.Node.EventType.TOUCH_END, this.retry, this);

        this.customManifestStr = (this.manifestUrl as any)._nativeAsset;

        console.log("manifestUrl:", this.manifestUrl);
        let version = JSON.parse((this.manifestUrl as any)._nativeAsset).version;
        console.log("version:", version);
        this.versionLab.string = version;

        this._storagePath = ((jsb.fileUtils ? jsb.fileUtils.getWritablePath() : '/') + 'blackjack-remote-asset');
        cc.log('Storage path for remote asset : ' + this._storagePath);

        // Setup your own version compare handler, versionA and B is versions in string
        // if the return value greater than 0, versionA is greater than B,
        // if the return value equals 0, versionA equals to B,
        // if the return value smaller than 0, versionA is smaller than B.
        
        this.versionCompareHandle = function (versionA, versionB) {
            cc.log("JS Custom Version Compare: version A is " + versionA + ', version B is ' + versionB);
            var vA = versionA.split('.');
            var vB = versionB.split('.');
            for (var i = 0; i < vA.length; ++i) {
                var a = parseInt(vA[i]);
                var b = parseInt(vB[i] || 0);
                if (a === b) {
                    continue;
                }
                else {
                    return a - b;
                }
            }
            if (vB.length > vA.length) {
                return -1;
            }
            else {
                return 0;
            }
        };
        

        // Init with empty manifest url for testing custom manifest
        this._am = new jsb.AssetsManager('', this._storagePath, this.versionCompareHandle);

        // Setup the verification callback, but we don't have md5 check function yet, so only print some message
        // Return true if the verification passed, otherwise return false
        this._am.setVerifyCallback(function (path, asset) {
            // When asset is compressed, we don't need to check its md5, because zip file have been deleted.
            var compressed = asset.compressed;
            // Retrieve the correct md5 value.
            var expectedMD5 = asset.md5;
            // asset.path is relative path and path is absolute.
            var relativePath = asset.path;
            // The size of asset file, but this value could be absent.
            var size = asset.size;
            if (compressed) {
                //this.msgLab.string = "Verification passed : " + relativePath;
                return true;
            }
            else {
                //this.msgLab.string = "Verification passed:" + relativePath + "(" + expectedMD5 + ")";
                return true;
            }
        });

        this.msgLab.string = 'Hot update is ready, please check or directly update.';

        if (cc.sys.os === cc.sys.OS_ANDROID) {
            // Some Android device may slow down the download process when concurrent tasks is too much.
            // The value may not be accurate, please do more test and find what's most suitable for your game.
            this._am.setMaxConcurrentTask(2);
            this.msgLab.string = "Max concurrent tasks count have been limited to 2";
        }
        
        this.fileProgress.progress = 0;
        this.byteProgress.progress = 0;
    }


    checkCb(event) {
        cc.log('Code: ' + event.getEventCode());
        switch (event.getEventCode())
        {
            case jsb.EventAssetsManager.ERROR_NO_LOCAL_MANIFEST:
                this.msgLab.string = "No local manifest file found, hot update skipped.";
                break;
            case jsb.EventAssetsManager.ERROR_DOWNLOAD_MANIFEST:
            case jsb.EventAssetsManager.ERROR_PARSE_MANIFEST:
                this.msgLab.string = "Fail to download manifest file, hot update skipped.";
                break;
            case jsb.EventAssetsManager.ALREADY_UP_TO_DATE:
                this.msgLab.string = "Already up to date with the latest remote version.";
                break;
            case jsb.EventAssetsManager.NEW_VERSION_FOUND:
                this.msgLab.string = 'New version found, please try to update. (' + this._am.getTotalBytes() + ')';
                this.checkBtn.active = false;
                this.fileProgress.progress = 0;
                this.byteProgress.progress = 0;
                break;
            default:
                return;
        }
        
        this._am.setEventCallback(null);
        this._checkListener = null;
        this._updating = false;
    }


    updateCb(event) {
        var needRestart = false;
        var failed = false;
        switch (event.getEventCode())
        {
            case jsb.EventAssetsManager.ERROR_NO_LOCAL_MANIFEST:
                this.msgLab.string = 'No local manifest file found, hot update skipped.';
                failed = true;
                break;
            case jsb.EventAssetsManager.UPDATE_PROGRESSION:
                this.byteProgress.progress = event.getPercent();
                this.fileProgress.progress = event.getPercentByFile();

                this.fileLabel.string = event.getDownloadedFiles() + ' / ' + event.getTotalFiles();
                this.byteLabel.string = event.getDownloadedBytes() + ' / ' + event.getTotalBytes();

                var msg = event.getMessage();
                if (msg) {
                    this.msgLab.string = 'Updated file: ' + msg;
                    // cc.log(event.getPercent()/100 + '% : ' + msg);
                }
                break;
            case jsb.EventAssetsManager.ERROR_DOWNLOAD_MANIFEST:
            case jsb.EventAssetsManager.ERROR_PARSE_MANIFEST:
                this.msgLab.string = 'Fail to download manifest file, hot update skipped.';
                failed = true;
                break;
            case jsb.EventAssetsManager.ALREADY_UP_TO_DATE:
                this.msgLab.string = 'Already up to date with the latest remote version.';
                failed = true;
                break;
            case jsb.EventAssetsManager.UPDATE_FINISHED:
                this.msgLab.string = 'Update finished. ' + event.getMessage();
                needRestart = true;
                break;
            case jsb.EventAssetsManager.UPDATE_FAILED:
                this.msgLab.string = 'Update failed. ' + event.getMessage();
                this.retryBtn.active = true;
                this._updating = false;
                this._canRetry = true;
                break;
            case jsb.EventAssetsManager.ERROR_UPDATING:
                this.msgLab.string = 'Asset update error: ' + event.getAssetId() + ', ' + event.getMessage();
                break;
            case jsb.EventAssetsManager.ERROR_DECOMPRESS:
                this.msgLab.string = event.getMessage();
                break;
            default:
                break;
        }

        if (failed) {
            this._am.setEventCallback(null);
            this._updateListener = null;
            this._updating = false;
        }

        if (needRestart) {
            console.log("needRestart");
            this._am.setEventCallback(null);
            this._updateListener = null;
            // Prepend the manifest's search path
            var searchPaths = jsb.fileUtils.getSearchPaths();
            var newPaths = this._am.getLocalManifest().getSearchPaths();
            console.log(JSON.stringify(newPaths));
            Array.prototype.unshift.apply(searchPaths, newPaths);
            // This value will be retrieved and appended to the default search path during game startup,
            // please refer to samples/js-tests/main.js for detailed usage.
            // !!! Re-add the search paths in main.js is very important, otherwise, new scripts won't take effect.
            cc.sys.localStorage.setItem('HotUpdateSearchPaths', JSON.stringify(searchPaths));
            jsb.fileUtils.setSearchPaths(searchPaths);

            cc.audioEngine.stopAll();
            cc.game.restart();
        }
    }

    loadCustomManifest () {
        if (this._am.getState() === jsb.AssetsManager.State.UNINITED) {
            var manifest = new jsb.Manifest(this.customManifestStr, this._storagePath);
            this._am.loadLocalManifest(manifest, this._storagePath);
            this.msgLab.string = 'Using custom manifest';
        }
    }
    
    retry () {
        if (!this._updating && this._canRetry) {
            this.retryBtn.active = false;
            this._canRetry = false;
            
            this.msgLab.string = 'Retry failed Assets...';
            this._am.downloadFailedAssets();
        }
    }
    
    checkUpdate () {
        if (this._updating) {
            this.msgLab.string = 'Checking or updating ...';
            return;
        }
        if (this._am.getState() === jsb.AssetsManager.State.UNINITED) {
            // Resolve md5 url
            var url = this.manifestUrl.nativeUrl;

            // if (cc.loader.md5Pipe) {
            //     url = cc.loader.md5Pipe.transformURL(url);
            // }


            // let uuid = (this.manifestUrl as any)._uuid;
            // url = cc.assetManager.utils.getUrlWithUuid(uuid, { isNative: true, ext: '.pem' });

            //url = url.replace('//', '/');

            console.log("manifestUrl 转换后:", url);

            this._am.loadLocalManifest(url);
        }

        console.log(this._am.getLocalManifest() , this._am.getLocalManifest().isLoaded())


        if (!this._am.getLocalManifest() || !this._am.getLocalManifest().isLoaded()) {
            this.msgLab.string = 'Failed to load local manifest ...';
            return;
        }
        this._am.setEventCallback(this.checkCb.bind(this));

        this._am.checkUpdate();
        this._updating = true;
    }

    hotUpdate () {
        if (this._am && !this._updating) {
            this._am.setEventCallback(this.updateCb.bind(this));

            if (this._am.getState() === jsb.AssetsManager.State.UNINITED) {
                // Resolve md5 url
                var url = this.manifestUrl.nativeUrl;
                // if (cc.loader.md5Pipe) {
                //     url = cc.loader.md5Pipe.transformURL(url);
                // }
                console.log("hotUpdate loadLocalManifest:",url);
                this._am.loadLocalManifest(url);
            }

            this._failCount = 0;
            this._am.update();
            this.updateBtn.active = false;
            this._updating = true;
            console.log("execute hotUpdate");
        }
    }
    
    show () {
        if (this.node.active === false) {
            this.node.active = true;
        }
    }

    onDestroy () {
        if (this._updateListener) {
            this._am.setEventCallback(null);
            this._updateListener = null;
        }
    }
}
